"""
Module for various Terraform backends.
"""

from typing import Union
from urllib.parse import urlparse

from infrahouse_toolkit.terraform.backends.exceptions import IHUnknownBackend
from infrahouse_toolkit.terraform.backends.s3backend import TFS3Backend


def get_backend(backend_id: str) -> Union[TFS3Backend,]:
    """
    Parse ``backend_id`` string and build a Terraform backend object based on the parsed values.

    :param backend_id: a straing that identifies a Terraform backend.
    :type backend_id: str
    :return: Terraform backend object.
    :rtype: Union[TFS3Backend, ]
    :raise IHUnknownBackend: If parsing failed or the encoded backend is not supported
    """
    result = urlparse(backend_id)
    if result.scheme == "s3":
        return TFS3Backend(result.netloc, result.path.lstrip("/"))
    raise IHUnknownBackend(f"Cannot find supported Terraform backend from {backend_id}.")



"""
.. topic:: ``ih-ec2``

    A group of commands to work with AWS EC2 service.

    See ``ih-ec2 --help`` for more details.
"""
import sys

import click
from boto3 import Session
from botocore.exceptions import (
    NoCredentialsError,
    NoRegionError,
    SSOTokenLoadError,
    TokenRetrievalError,
)

from infrahouse_toolkit import LOG
from infrahouse_toolkit.aws import aws_sso_login
from infrahouse_toolkit.aws.config import AWSConfig
from infrahouse_toolkit.cli.ih_ec2.cmd_instance_types import cmd_instance_types
from infrahouse_toolkit.cli.ih_ec2.cmd_launch import cmd_launch
from infrahouse_toolkit.cli.ih_ec2.cmd_launch_templates import cmd_launch_templates
from infrahouse_toolkit.cli.ih_ec2.cmd_list import cmd_list
from infrahouse_toolkit.cli.ih_ec2.cmd_subnets import cmd_subnets
from infrahouse_toolkit.cli.ih_ec2.cmd_terminate import cmd_terminate
from infrahouse_toolkit.logging import setup_logging

AWS_DEFAULT_REGION = "us-west-1"


def get_aws_regions() -> list:
    """
    :return: List of AWS regions.
    :rtype: list
    """
    return [
        "af-south-1",
        "ap-east-1",
        "ap-northeast-1",
        "ap-northeast-2",
        "ap-northeast-3",
        "ap-south-1",
        "ap-southeast-1",
        "ap-southeast-2",
        "ap-southeast-3",
        "ca-central-1",
        "eu-central-1",
        "eu-north-1",
        "eu-south-1",
        "eu-west-1",
        "eu-west-2",
        "eu-west-3",
        "me-south-1",
        "sa-east-1",
        "us-east-1",
        "us-east-2",
        "us-west-1",
        "us-west-2",
    ]


@click.group()
@click.option(
    "--debug",
    help="Enable debug logging.",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "--aws-profile",
    help="AWS profile name for authentication.",
    type=click.Choice(AWSConfig().profiles),
    default=None,
    show_default=True,
)
@click.option(
    "--aws-region",
    help="AWS region to use.",
    type=click.Choice(get_aws_regions()),
    show_default=True,
    default=None,
)
@click.version_option()
@click.pass_context
def ih_ec2(ctx, **kwargs):
    """AWS EC2 helpers."""
    setup_logging(LOG, debug=kwargs["debug"])
    aws_profile = kwargs["aws_profile"]
    aws_config = AWSConfig()
    aws_region = kwargs["aws_region"]
    aws_session = None
    try:
        response = get_aws_client("sts", aws_profile, aws_region).get_caller_identity()

        LOG.info("Connected to AWS as %s", response["Arn"])
    except NoCredentialsError as err:
        LOG.error(err)
        LOG.info("Try to run ih-ec2 with --aws-profile option.")
        LOG.info("Available profiles:\n\t%s", "\n\t".join(aws_config.profiles))
        sys.exit(1)

    except (SSOTokenLoadError, TokenRetrievalError) as err:
        if not aws_profile:
            LOG.info("Try to run ih-ec2 with --aws-profile option.")
            LOG.info("Available profiles:\n\t%s", "\n\t".join(aws_config.profiles))
            sys.exit(1)
        LOG.debug(err)
        aws_session = aws_sso_login(aws_config, aws_profile, region=aws_region)
        response = get_aws_client("sts", aws_profile, aws_region, session=aws_session).get_caller_identity()

        LOG.info("Connected to AWS as %s", response["Arn"])

    try:
        ctx.obj = {
            "debug": kwargs["debug"],
            "ec2_client": get_aws_client("ec2", aws_profile, aws_region, session=aws_session),
            "aws_config": aws_config,
        }
    except NoRegionError as err:
        LOG.error(err)
        LOG.error("Use the --aws-region option to specify the AWS region.")
        sys.exit(1)


for cmd in [cmd_launch, cmd_list, cmd_instance_types, cmd_terminate, cmd_subnets, cmd_launch_templates]:
    # noinspection PyTypeChecker
    ih_ec2.add_command(cmd)


def get_aws_client(service_name: str, profile: str, region: str, session=None):
    """
    Get a client instance for an AWS service.

    :param service_name: AWS service e.g. ``ec2``.
    :param profile: AWS profile for authentication.
    :param region: AWS region.
    :param session: if an AWS session is passed, use it to create a client.
    :type session: Session
    :return: A client instance.
    """
    session = session or Session(region_name=region, profile_name=profile)
    return session.client(service_name)



"""
Top-level package for InfraHouse Toolkit.

:Author: Oleksandr Kuzminskyi
:Version: 2.9.4
:Email: aleks@infrahouse.com
"""

__author__ = """Oleksandr Kuzminskyi"""
__email__ = "aleks@infrahouse.com"
__version__ = "2.9.4"

from logging import getLogger

DEFAULT_OPEN_ENCODING = "utf8"
DEFAULT_ENCODING = DEFAULT_OPEN_ENCODING
LOG = getLogger(__name__)



"""
AWS classes.
"""
import time
import webbrowser
from os import path as osp
from pprint import pformat
from time import sleep

from boto3 import Session
from diskcache import Cache

from infrahouse_toolkit import LOG
from infrahouse_toolkit.aws.config import AWSConfig
from infrahouse_toolkit.aws.exceptions import IHAWSException


def aws_sso_login(aws_config: AWSConfig, profile_name: str, region: str = None):
    """
    Login into AWS using SSO.

    Credit:
    https://stackoverflow.com/questions/62311866/how-to-use-the-aws-python-sdk-while-connecting-via-sso-credentials
    """
    cache_directory = osp.expanduser("~/.infrahouse-toolkit")
    with Cache(directory=cache_directory) as cache_reference:
        cache_key = f"ih-ec2-credentials-{profile_name}"
        credentials = cache_reference.get(cache_key)
        if not credentials:
            credentials = _get_credentials(aws_config, profile_name)
            cache_reference.set(
                cache_key, credentials, expire=int(int(credentials["expiration"]) / 1000 - int(time.time()))
            )

    return Session(
        region_name=region or aws_config.get_region(profile_name),
        aws_access_key_id=credentials["accessKeyId"],
        aws_secret_access_key=credentials["secretAccessKey"],
        aws_session_token=credentials["sessionToken"],
    )


def _get_credentials(aws_config: AWSConfig, profile_name: str):
    """
    Login into AWS using SSO.

    Credit:
    https://stackoverflow.com/questions/62311866/how-to-use-the-aws-python-sdk-while-connecting-via-sso-credentials

    :raise IHAWSException: If user didn't confirm auth
    """

    session = Session()
    sso_oidc = session.client("sso-oidc")
    client_creds = sso_oidc.register_client(
        clientName="infrahouse-toolkit",
        clientType="public",
    )
    LOG.debug("client_creds = %s", pformat(client_creds, indent=4))
    device_authorization = sso_oidc.start_device_authorization(
        clientId=client_creds["clientId"],
        clientSecret=client_creds["clientSecret"],
        startUrl=aws_config.get_start_url(profile_name),
    )

    LOG.debug("device_authorization = %s", pformat(device_authorization, indent=4))
    url = device_authorization["verificationUriComplete"]
    device_code = device_authorization["deviceCode"]
    expires_in = device_authorization["expiresIn"]
    interval = device_authorization["interval"]
    LOG.info("Verify user code: %s", device_authorization["userCode"])
    webbrowser.open(url, autoraise=True)
    for _ in range(1, expires_in // interval + 1):
        sleep(interval)
        try:
            token = sso_oidc.create_token(
                grantType="urn:ietf:params:oauth:grant-type:device_code",
                deviceCode=device_code,
                clientId=client_creds["clientId"],
                clientSecret=client_creds["clientSecret"],
            )
            access_token = token["accessToken"]
            sso = session.client("sso")
            role_creds = sso.get_role_credentials(
                roleName=aws_config.get_role(profile_name),
                accountId=aws_config.get_account_id(profile_name),
                accessToken=access_token,
            )["roleCredentials"]
            return role_creds
        except sso_oidc.exceptions.AuthorizationPendingException:
            pass

    raise IHAWSException(f"The verification code isn't confirmed by user in {expires_in} seconds")



"""
.. topic:: ``ih-plan upload``

    A ``ih-plan upload`` subcommand.

    See ``ih-plan upload --help`` for more details.
"""

from os import path as osp

import click

from infrahouse_toolkit.cli.lib import get_bucket, get_s3_client


@click.command(name="upload")
@click.option("--key-name", help="Path to the file in the S3 bucket. Default is pending/<plan_file>", default=None)
@click.argument("plan_file")
@click.pass_context
def cmd_upload(ctx, key_name, plan_file):
    """
    Upload a plan file to an S3 bucket.

    The specified plan file will be uploaded to the S3 bucket (See ih-plan --help)
    and will be available as

    s3://<bucket name>/<key_name>

    By default, the S3 url will be s3://<bucket name>/pending/<plan_file>
    """
    s3_client = get_s3_client(role=ctx.obj["aws_assume_role_arn"])
    bucket = ctx.obj["bucket"] or get_bucket(ctx.obj["tf_backend_file"])
    dst_name = key_name or osp.join("pending", plan_file)
    s3_client.upload_file(plan_file, bucket, dst_name)
    print(f"Successfully uploaded s3://{bucket}/{dst_name}.")



from textwrap import dedent

import pytest

from infrahouse_toolkit.aws import AWSConfig

CONTENT = dedent(
    """
    [default]
    sso_account_id = 123

    [profile foo]
    sso_account_id = 456
    """
)


@pytest.mark.parametrize(
    "profile, account_id",
    [
        (
            "default",
            "123",
        ),
        (
            "foo",
            "456",
        ),
    ],
)
def test_account_id(profile, account_id, tmpdir):
    aws_home = tmpdir.mkdir("home")
    cfg = aws_home.join("config")
    cfg.write(CONTENT)
    assert AWSConfig(aws_home=str(aws_home)).get_account_id(profile) == account_id



"""
.. topic:: ``ih-puppet apply``

    A ``ih-puppet apply`` subcommand.

    See ``ih-puppet apply`` for more details.
"""
import json
import os
import re
import sys
from os import environ
from os import path as osp
from subprocess import PIPE, Popen

import click

from infrahouse_toolkit import DEFAULT_OPEN_ENCODING, LOG
from infrahouse_toolkit.lock.system import SystemLock


@click.command(name="apply")
@click.argument("manifest", required=False)
@click.pass_context
def cmd_apply(ctx, manifest):
    """
    Apply puppet manifest
    """
    manifest = manifest or f"{ctx.obj['root_directory']}/environments/{ctx.obj['environment']}/manifests/site.pp"
    LOG.info("Applying puppet manifest in %s", manifest)
    cmd = ["puppet", "apply"]
    if ctx.obj["debug"]:
        cmd.append("-d")

    cmd.extend(
        [
            "--environment",
            ctx.obj["environment"],
            "--environmentpath",
            ctx.obj["environmentpath"],
            "--hiera_config",
            ctx.obj["hiera_config"],
            f"--modulepath={ctx.obj['module_path']}",
            manifest,
            "--write-catalog-summary",
            "--detailed-exitcodes",
        ]
    )

    with SystemLock("/var/run/ih-puppet-apply.lock"):
        env = {"PATH": f"{environ['PATH']}:/opt/puppetlabs/bin"}
        install_module_dependencies(module_path=ctx.obj["module_path"], env=env)
        LOG.debug("Executing %s", " ".join(cmd))
        # First run is to update the puppet code
        with Popen(cmd, env=env) as proc:
            proc.communicate()

        # Second run is to apply whatever the new puppet code brings
        with Popen(cmd, env=env) as proc:
            proc.communicate()
            ret = proc.returncode
            LOG.debug("Exit code: %d", ret)
            if ret == 0:
                LOG.info("The run succeeded with no changes or failures; the system was already in the desired state.")
                sys.exit(0)
            elif ret == 1:
                LOG.error("The run failed.")
                sys.exit(ret)
            elif ret == 2:
                LOG.info("The run succeeded, and some resources were changed.")
                sys.exit(0)
            elif ret == 4:
                LOG.warning("The run succeeded, and some resources failed.")
                sys.exit(ret)
            elif ret == 6:
                LOG.warning("The run succeeded, and included both changes and failures.")
                sys.exit(ret)
            else:
                LOG.error("Unknown run state.")
                sys.exit(ret)


def install_module_dependencies(module_path: str, env: dict = None):
    """
    Assuming each subdirectory in ``module_path`` is a puppet module,
    read its ``metadata.json`` and install puppet module dependencies
    in the same directory.

    :param module_path: Path to a directory with puppet modules.
    :type module_path: str
    :param env: Environment variables for the puppet command.
    :type env: dict
    """
    for module in os.listdir(module_path):
        LOG.info("Installing %s dependencies", module)
        try:
            with open(osp.join(module_path, module, "metadata.json"), encoding=DEFAULT_OPEN_ENCODING) as f_desc:
                deps = json.loads(f_desc.read())["dependencies"]
                for dep in deps:
                    cmd = [
                        "puppet",
                        "module",
                        "--render-as",
                        "json",
                        "--modulepath",
                        module_path,
                        "install",
                        dep["name"],
                        "-v",
                        dep["version_requirement"],
                    ]
                    kwargs = {"stdout": PIPE, "stderr": PIPE}
                    if env:
                        kwargs["env"] = env

                    with Popen(cmd, **kwargs) as proc:
                        cout, cerr = proc.communicate()
                        LOG.info("STDOUT: \n%s", strip_colors(cout.decode()))
                        ret = proc.returncode
                        LOG.debug("Exit code: %d", ret)
                        if ret != 0:
                            LOG.error("Command '%s' exited with %d.", " ".join(cmd), ret)
                            LOG.error("STDERR:\n%s", strip_colors(cerr.decode()))
                            sys.exit(1)
        except NotADirectoryError:
            LOG.info("%s isn't a puppet module")


def strip_colors(text: str) -> str:
    """
    Credit:
    https://stackoverflow.com/questions/14693701/how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python

    :param text: ANSI colored text
    :return: The input text w/o colors.
    """
    ansi_escape = re.compile(
        r"""
    \x1B  # ESC
    (?:   # 7-bit C1 Fe (except CSI)
        [@-Z\\-_]
    |     # or [ for CSI, followed by a control sequence
        \[
        [0-?]*  # Parameter bytes
        [ -/]*  # Intermediate bytes
        [@-~]   # Final byte
    )
""",
        re.VERBOSE,
    )
    return ansi_escape.sub("", text)



from textwrap import dedent

from click.testing import CliRunner

from infrahouse_toolkit.cli.ih_plan import ih_plan


def test_help():
    runner = CliRunner()
    # noinspection PyTypeChecker
    result = runner.invoke(ih_plan, ["min-permissions", "--help"])
    assert result.exit_code == 0
    assert "Usage: ih-plan min-permissions" in result.output


def test_simple(tmpdir):
    tracefile = tmpdir.join("trace")
    tracefile.write(
        dedent(
            """
            {"aws.operation": "PutScalingPolicy","aws.service": "Auto Scaling"}
            {"aws.operation": "DescribePolicies","aws.service": "Auto Scaling"}
            """
        )
    )

    runner = CliRunner()
    # noinspection PyTypeChecker
    result = runner.invoke(ih_plan, ["min-permissions", str(tracefile)])
    assert result.exit_code == 0
    print(result.output)
    assert (
        result.output
        == """## Existing 0 actions:
[]
## 2 new action(s):
[
    "autoscaling:DescribePolicies",
    "autoscaling:PutScalingPolicy"
]
## Old and new actions together excluding duplicates, 2 in total:
[
    "autoscaling:DescribePolicies",
    "autoscaling:PutScalingPolicy"
]
"""
    )


def test_simple_existing(tmpdir):
    tracefile = tmpdir.join("trace")
    tracefile.write(
        dedent(
            """
                {"tf_resource_type": "aws_s3_bucket_versioning","tf_rpc": "ApplyResourceChange"}
                {"tf_resource_type": "aws_s3_bucket_versioning","tf_rpc": "ApplyResourceChange"}
            """
        )
    )

    existing_actions = tmpdir.join("existing.json")
    existing_actions.write(
        dedent(
            """
            ["s3:PutBucketVersioning"]
            """
        )
    )

    runner = CliRunner()
    # noinspection PyTypeChecker
    result = runner.invoke(ih_plan, ["min-permissions", "--existing-actions", str(existing_actions), str(tracefile)])
    assert result.exit_code == 0
    print(result.output)
    assert (
        result.output
        == """## Existing 1 actions:
[
    "s3:PutBucketVersioning"
]
## 0 new action(s):
[]
## Old and new actions together excluding duplicates, 1 in total:
[
    "s3:PutBucketVersioning"
]
"""
    )



from os import environ

import pytest

from infrahouse_toolkit.terraform import parse_comment, parse_plan
from infrahouse_toolkit.terraform.backends import TFS3Backend
from infrahouse_toolkit.terraform.githubpr import GitHubPR


@pytest.mark.skipif("GITHUB_TOKEN" not in environ, reason="This is a development test, needs GITHUB_TOKEN")
def test_find_comment_by_backend():
    gh_pr = GitHubPR("infrahouse8/github-control", 33)
    comment = gh_pr.find_comment_by_backend(TFS3Backend("foo-bucket", "path/to/key.state"))
    if comment:
        status = parse_comment(comment.body)
        counts, resources = parse_plan(open("infrahouse_toolkit/terraform/tests/plans/plan-2-1-2.stdout").read())
        status.affected_resources = resources
        status.add = counts.add
        status.change = counts.change
        status.destroy = counts.destroy
        status.success = False
        print(status.comment)

        comment.edit(status.comment)



"""
InfraHouse Toolkit timeout module.
"""
import signal
from contextlib import contextmanager


@contextmanager
def timeout(seconds: int):
    """
    Timeout context manager.

    :param seconds: Max execution time in seconds.
    :type seconds: int
    :raise TimeoutError: when the code under a ``with`` is running
        more than ``seconds``.
    """

    def handler(signum, frame):
        if signum or frame:
            pass
        raise TimeoutError(f"Executing timed out after {seconds} seconds")

    original_handler = signal.signal(signal.SIGALRM, handler)
    try:
        signal.alarm(seconds)
        yield
    finally:
        signal.alarm(0)
        signal.signal(signal.SIGALRM, original_handler)



"""
.. topic:: ``ih-plan min-permissions``

    A ``ih-plan min-permissions`` subcommand.

    See ``ih-plan min-permissions --help`` for more details.
"""
import json
from json import JSONDecodeError

import click

from infrahouse_toolkit import DEFAULT_OPEN_ENCODING


@click.command(name="min-permissions")
@click.option("--existing-actions", help="A file with permissions.", default=None)
@click.argument("trace_file")
def cmd_min_permissions(existing_actions, trace_file):
    """
    Parse Terraform trace file and produce an action list from the trace.

    The trace file contains entries with AWS actions. The command
    finds the actions and AWS services to generate a list that
    you can add to an AWS policy.
    It's useful to prepare the least privileges policy.

    The output looks similar to this:

    \b
    [
        "ec2:DeleteNatGateway",
        "ec2:DescribeAddresses",
        "ec2:DescribeInternetGateways",
        "ec2:DescribeNatGateways",
    ]

    """
    actions = ActionList()
    if existing_actions:
        actions.load_from_file(existing_actions)

    print(f"## Existing {actions.count} actions:")
    print(actions)

    new_actions = ActionList()
    new_actions.parse_trace(trace_file, existing=actions.actions)
    print(f"## {new_actions.count} new action(s):")
    print(str(new_actions))

    combined_actions = ActionList()
    for action in actions.actions + new_actions.actions:
        combined_actions.add(action)

    print(f"## Old and new actions together excluding duplicates, {combined_actions.count} in total:")
    print(str(combined_actions))


class ActionList:
    """
    List of AWS actions. Action here is a string as in AWS's policy e.g. ``ec2:DescribeInstances``.

    """

    SERVICE_NAMING_MAP = {
        "auto scaling": "autoscaling",
        "elastic load balancing v2": "elasticloadbalancing",
        "route 53": "route53",
        "secrets manager": "secretsmanager",
        "cloudwatch logs": "logs",
        "eventbridge": "events",
    }
    PERMISSION_NAMING_MAP = {
        "HeadBucket": "ListBucket",
        "HeadObject": "GetObject",
        "GetBucketAccelerateConfiguration": "GetAccelerateConfiguration",
        "GetBucketEncryption": "GetEncryptionConfiguration",
        "GetBucketCors": "GetBucketCORS",
        "GetBucketLifecycleConfiguration": "GetLifecycleConfiguration",
        "GetBucketReplication": "GetReplicationConfiguration",
        "GetObjectLockConfiguration": "GetBucketObjectLockConfiguration",
        "DeletePublicAccessBlock": "PutBucketPublicAccessBlock",
        "GetPublicAccessBlock": "GetBucketPublicAccessBlock",
        "PutPublicAccessBlock": "PutBucketPublicAccessBlock",
    }
    # Some permissions require additional ones.
    REQUIRED_EXTRA_PERMISSIONS_MAP = {
        "autoscaling:CreateAutoScalingGroup": ["iam:PassRole", "iam:CreateServiceLinkedRole"],
        "autoscaling:UpdateAutoScalingGroup": ["iam:PassRole"],
        "elasticloadbalancing:CreateLoadBalancer": ["elasticloadbalancing:AddTags"],
        "iam:AddRoleToInstanceProfile": ["iam:PassRole"],
        "ec2:CreateLaunchTemplate": ["ec2:CreateTags"],
        "ec2:ImportKeyPair": ["ec2:CreateTags"],
        "ec2:RunInstances": ["ec2:CreateTags"],
    }

    def __init__(self):
        self._actions = set()

    @property
    def actions(self) -> list:
        """List of action strings."""
        return sorted(list(self._actions))

    @property
    def count(self) -> int:
        """Number of actions in the list."""
        return len(self._actions)

    def add(self, action: str):
        """Add a new action. Convert service name to the AWS policy format and add dependent actions if any."""
        norm_action = self._normalize_action(action)
        self._actions.add(norm_action)
        for dependency in self.REQUIRED_EXTRA_PERMISSIONS_MAP.get(str(norm_action), []):
            self._actions.add(dependency)

    def load_from_file(self, file):
        """Load actions from a file with a JSON. The JSON should be an array of strings."""
        with open(file, encoding=DEFAULT_OPEN_ENCODING) as f_desc:
            for action in json.loads(f_desc.read()):
                self.add(action)

    def parse_trace(self, file, existing=None):
        """Inspect a Terraform trace file and collect actions"""
        existing_permissions = existing or []
        with open(file, encoding=DEFAULT_OPEN_ENCODING) as f_decs:
            for line in f_decs.readlines():
                try:
                    operation = json.loads(line)
                    operation_key = "aws.operation" if "aws.operation" in operation else "rpc.method"
                    service_key = "aws.service" if "aws.service" in operation else "rpc.service"
                    if all((operation_key in operation, service_key in operation)):
                        service_name = operation[service_key].lower()
                        permission = self._normalize_action(f"{service_name}:{operation[operation_key]}")
                        if permission not in existing_permissions:
                            self.add(permission)
                    elif all(
                        (
                            operation.get("tf_resource_type") == "aws_s3_bucket_versioning",
                            operation.get("tf_rpc") == "ApplyResourceChange",
                        )
                    ):
                        permission = "s3:PutBucketVersioning"
                    elif all(
                        (
                            operation.get("tf_resource_type") == "aws_s3_bucket_server_side_encryption_configuration",
                            operation.get("tf_rpc") == "ApplyResourceChange",
                        )
                    ):
                        permission = "s3:PutEncryptionConfiguration"
                    else:
                        continue

                    if permission not in existing_permissions:
                        self.add(permission)

                except JSONDecodeError:
                    pass

    def _normalize_action(self, action):
        if ":" in action:
            s_part, a_part = action.split(":")
            return f"{self.SERVICE_NAMING_MAP.get(s_part, s_part)}:{self.PERMISSION_NAMING_MAP.get(a_part, a_part)}"
        return action

    def __str__(self):
        return json.dumps(self.actions, indent=4)



from infrahouse_toolkit.terraform.backends import TFS3Backend
from infrahouse_toolkit.terraform.status import RunOutput, RunResult, TFStatus


def test_comment():
    status = TFStatus(
        TFS3Backend("foo_backet", "path/to/tf.state"), True, RunResult(1, 1, 1), RunOutput("no stdout", None)
    )
    assert isinstance(status.comment, str)
    print(status.comment)


def test_comment_none():
    status = TFStatus(
        TFS3Backend("foo_backet", "path/to/tf.state"),
        True,
        RunResult(None, None, None),
        RunOutput("no stdout", "no stderr"),
        affected_resources=RunResult(None, None, None),
    )
    assert isinstance(status.comment, str)
    print(status.comment)
